@@ -1,11 +1,11 @@
 use strict;
 use warnings;
-package JSON::Schema::Modern::Document::OpenAPI; # git description: v0.007-5-g4c4daef
+package JSON::Schema::Modern::Document::OpenAPI; # git description: v0.008-16-g34b098c
 # vim: set ts=8 sts=2 sw=2 tw=100 et :
 # ABSTRACT: One OpenAPI v3.1 document
 # KEYWORDS: JSON Schema data validation request response OpenAPI
 
-our $VERSION = '0.008';
+our $VERSION = '0.009';
 
 use 5.020;  # for fc, unicode_strings features
 use Moo;
@@ -157,7 +157,7 @@ sub traverse ($self, $evaluator) {
     my ($operation_id, $path) = @$pair;
     if (my $existing = $self->get_operationId($operation_id)) {
       ()= E({ %$state, keyword => 'operationId', schema_path => $path },
-        'duplicate operationId (with location %s)', $existing);
+        'duplicate of operationId at %s', $existing);
     }
     else {
       $self->_add_operationId($operation_id => $path);
@@ -213,7 +213,7 @@ JSON::Schema::Modern::Document::OpenAPI - One OpenAPI v3.1 document
 
 =head1 VERSION
 
-version 0.008
+version 0.009
 
 =head1 SYNOPSIS
 
