@@ -5,7 +5,7 @@ package OpenAPI::Modern;
 # ABSTRACT: Validate HTTP requests and responses against an OpenAPI document
 # KEYWORDS: validation evaluation JSON Schema OpenAPI Swagger HTTP request response
 
-our $VERSION = '0.008';
+our $VERSION = '0.009';
 
 use 5.020;  # for fc, unicode_strings features
 use Moo;
@@ -21,8 +21,8 @@ use Ref::Util qw(is_plain_hashref is_plain_arrayref is_ref);
 use List::Util 'first';
 use Scalar::Util 'looks_like_number';
 use Feature::Compat::Try;
-use JSON::Schema::Modern 0.527;
-use JSON::Schema::Modern::Utilities qw(jsonp canonical_uri E abort);
+use JSON::Schema::Modern 0.531;
+use JSON::Schema::Modern::Utilities 0.531 qw(jsonp unjsonp canonical_uri E abort);
 use JSON::Schema::Modern::Document::OpenAPI;
 use MooX::HandlesVia;
 use MooX::TypeTiny 0.002002;
@@ -75,7 +75,7 @@ around BUILDARGS => sub ($orig, $class, @args) {
 };
 
 # at the moment, we rely on these values being provided in $options:
-# - path_template
+# - path_template OR operationId
 # - path_captures
 sub validate_request ($self, $request, $options) {
   my $state = {
@@ -86,23 +86,19 @@ sub validate_request ($self, $request, $options) {
     errors => [],
   };
 
-  my $path_template = $options->{path_template};
-  my $captures = $options->{path_captures};
-  croak 'missing parameter path_template' if not length $path_template;
-  croak 'missing parameter captures' if not is_plain_hashref($captures);
+  croak 'missing option path_template or operation_id'
+    if not exists $options->{path_template} and not exists $options->{operation_id};
+
+  my $path_captures = $options->{path_captures};
+  croak 'missing option path_captures' if not is_plain_hashref($path_captures);
 
   try {
+    my $path_template = $self->_find_path($state, $request, $options);
+    my $path_item = $self->openapi_document->schema->{paths}{$path_template};
     my $method = lc $request->method;
-    my $schema = $self->openapi_document->schema;
-
-    my $path_item = $schema->{paths}{$path_template};
-    abort({ %$state, keyword => 'paths', _schema_path_suffix => $path_template },
-      'missing path-item "%s"', $path_template) if not $path_item;
+    my $operation = $path_item->{$method};
 
     $state->{schema_path} = jsonp('/paths', $path_template);
-    my $operation = $path_item->{$method};
-    abort({ %$state, keyword => $method }, 'missing entry for HTTP method "%s"', $method)
-      if not $operation;
 
     # PARAMETERS
     # { $in => { $name => 'path-item'|$method } }  as we process each one.
@@ -128,7 +124,7 @@ sub validate_request ($self, $request, $options) {
 
         $state->{data_path} = jsonp($state->{data_path}, $param_obj->{in}, $param_obj->{name});
         my $valid =
-            $param_obj->{in} eq 'path' ? $self->_validate_path_parameter($state, $param_obj, $captures)
+            $param_obj->{in} eq 'path' ? $self->_validate_path_parameter($state, $param_obj, $path_captures)
           : $param_obj->{in} eq 'query' ? $self->_validate_query_parameter($state, $param_obj, $request->uri)
           : $param_obj->{in} eq 'header' ? $self->_validate_header_parameter($state, $param_obj->{name}, $param_obj, $request->headers)
           : $param_obj->{in} eq 'cookie' ? $self->_validate_cookie_parameter($state, $param_obj, $request)
@@ -160,7 +156,7 @@ sub validate_request ($self, $request, $options) {
       push @{$state->{errors}}, $e;
     }
     else {
-      E($state, 'EXCEPTION: '.$e);
+      ()= E($state, 'EXCEPTION: '.$e);
     }
   }
 
@@ -168,7 +164,7 @@ sub validate_request ($self, $request, $options) {
 }
 
 # at the moment, we rely on these values being provided in $options:
-# - path_template
+# - path_template OR operationId
 sub validate_response ($self, $response, $options) {
   my $state = {
     data_path => '/response',
@@ -178,24 +174,17 @@ sub validate_response ($self, $response, $options) {
     errors => [],
   };
 
-  my $path_template = $options->{path_template};
-  croak 'missing parameter path_template' if not length $path_template;
+  croak 'missing option path_template or operation_id'
+    if not exists $options->{path_template} and not exists $options->{operation_id};
 
   try {
-    my $schema = $self->openapi_document->schema;
-    abort({ %$state, keyword => 'paths', _schema_path_suffix => $path_template },
-        'missing path-item "%s"', $path_template)
-      if not $schema->{paths}{$path_template};
-
-    $state->{schema_path} = jsonp('/paths', $path_template);
+    my $path_template = $self->_find_path($state, $response->request, $options);
     my $method = lc $response->request->method;
-    my $operation = $schema->{paths}{$path_template}{$method};
-    abort({ %$state, _schema_path_suffix => $method }, 'missing operation')
-      if not $operation;
+    my $operation = $self->openapi_document->schema->{paths}{$path_template}{$method};
 
     return $self->_result($state) if not exists $operation->{responses};
 
-    $state->{schema_path} = jsonp($state->{schema_path}, $method);
+    $state->{schema_path} = jsonp('/paths', $path_template, $method);
 
     my $response_property = first { exists $operation->{responses}{$_} }
       $response->code, substr(sprintf('%03s', $response->code), 0, -2).'XX', 'default';
@@ -236,7 +225,7 @@ sub validate_response ($self, $response, $options) {
       push @{$state->{errors}}, $e;
     }
     else {
-      E($state, 'EXCEPTION: '.$e);
+      ()= E($state, 'EXCEPTION: '.$e);
     }
   }
 
@@ -245,13 +234,57 @@ sub validate_response ($self, $response, $options) {
 
 ######## NO PUBLIC INTERFACES FOLLOW THIS POINT ########
 
+# at the moment, we rely on these values being provided in $options:
+# - path_template OR operation_id
+# return: path_template.  caller can get method out of $request.
+# in the future, we will parse the request URI and can extract path_captures as well.
+sub _find_path ($self, $state, $request, $options) {
+  my $path_template;
+
+  # path_template from options, method from request
+  if (exists $options->{path_template}) {
+    $path_template = $options->{path_template};
+
+    my $path_item = $self->openapi_document->schema->{paths}{$path_template};
+    abort({ %$state, keyword => 'paths', _schema_path_suffix => $path_template },
+      'missing path-item "%s"', $path_template) if not $path_item;
+
+    my $method = lc $request->method;
+    abort({ %$state, schema_path => jsonp('/paths', $path_template), keyword => $method },
+        'missing entry for HTTP method "%s"', $method)
+      if not $path_item->{$method};
+  }
+
+  # path_template and method from operationId from options
+  if (exists $options->{operation_id}) {
+    my $operation_path = $self->openapi_document->get_operationId($options->{operation_id});
+    abort({ %$state, keyword => 'paths' }, 'unknown operation_id "%s"', $options->{operation_id})
+      if not $operation_path;
+    abort({ %$state, schema_path => $operation_path, keyword => 'operationId' },
+      'operation id does not have an associated path') if $operation_path !~ m{^/paths/};
+    (undef, undef, $path_template, my $method) = unjsonp($operation_path);
+
+    abort({ %$state, schema_path => jsonp('/paths', $path_template) },
+        'operation does not match provided path_template')
+      if exists $options->{path_template} and $options->{path_template} ne $path_template;
+
+    abort({ %$state, schema_path => $operation_path }, 'wrong HTTP method %s', $request->method)
+      if lc $request->method ne $method;
+  }
+
+  # TODO: alternatively, look up $path_template and $path_captures from $request->uri,
+  # OR verify that $path_template matches $request->uri
+
+  return $path_template;
+}
+
 # for now, we only use captures, rather than parsing the URI directly.
-sub _validate_path_parameter ($self, $state, $param_obj, $captures) {
+sub _validate_path_parameter ($self, $state, $param_obj, $path_captures) {
   # 'required' is always true for path parameters
   return E({ %$state, keyword => 'required' }, 'missing path parameter: %s', $param_obj->{name})
-    if not exists $captures->{$param_obj->{name}};
+    if not exists $path_captures->{$param_obj->{name}};
 
-  $self->_validate_parameter_content($state, $param_obj, \ $captures->{$param_obj->{name}});
+  $self->_validate_parameter_content($state, $param_obj, \ $path_captures->{$param_obj->{name}});
 }
 
 sub _validate_query_parameter ($self, $state, $param_obj, $uri) {
@@ -293,23 +326,30 @@ sub _validate_cookie_parameter ($self, $state, $param_obj, $request) {
 
 sub _validate_parameter_content ($self, $state, $param_obj, $content_ref) {
   if (exists $param_obj->{content}) {
-    my ($content_type) = keys $param_obj->{content}->%*;
+    abort({ %$state, keyword => 'content' }, 'more than one media type entry present')
+      if keys $param_obj->{content}->%* > 1;  # TODO: remove, when the spec schema is updated
+    my ($media_type) = keys $param_obj->{content}->%*;  # there can only be one key
+    my $schema = $param_obj->{content}{$media_type}{schema};
 
-    my $media_type_decoder = $self->get_media_type($content_type);
-    abort({ %$state, keyword => 'content', _schema_path_suffix => $content_type },
-        'EXCEPTION: unsupported Content-Type "%s": add support with $openapi->add_media_type(...)', $content_type)
-      if not $media_type_decoder;
+    my $media_type_decoder = $self->get_media_type($media_type);  # case-insensitive, wildcard lookup
+    if (not $media_type_decoder) {
+      # don't fail if the schema would pass on any input
+      return if is_plain_hashref($schema) ? !keys %$schema : $schema;
+
+      abort({ %$state, keyword => 'content', _schema_path_suffix => $media_type},
+        'EXCEPTION: unsupported media type "%s": add support with $openapi->add_media_type(...)', $media_type)
+    }
 
     try {
       $content_ref = $media_type_decoder->($content_ref);
     }
     catch ($e) {
-      return E({ %$state, keyword => 'content', _schema_path_suffix => $content_type },
-        'could not decode content as %s', $content_type);
+      return E({ %$state, keyword => 'content', _schema_path_suffix => $media_type },
+        'could not decode content as %s', $media_type);
     }
 
-    $state = { %$state, schema_path => jsonp($state->{schema_path}, 'content', $content_type, 'schema') };
-    return $self->_evaluate_subschema($content_ref->$*, $param_obj->{content}{$content_type}{schema}, $state);
+    $state = { %$state, schema_path => jsonp($state->{schema_path}, 'content', $media_type, 'schema') };
+    return $self->_evaluate_subschema($content_ref->$*, $schema, $state);
   }
 
   $state = { %$state, schema_path => jsonp($state->{schema_path}, 'schema') };
@@ -317,23 +357,28 @@ sub _validate_parameter_content ($self, $state, $param_obj, $content_ref) {
 }
 
 sub _validate_body_content ($self, $state, $content_obj, $message) {
-  my $content_type = $message->content_type;
+  my $content_type = fc $message->content_type;
 
   return E({ %$state, data_path => $state->{data_path} =~ s{body}{header/Content-Type}r, keyword => 'content' },
       'missing header: Content-Type')
     if not length $content_type;
 
-  # TODO: respect wildcard entries in the openapi document
-  return E({ %$state, keyword => 'content' }, 'incorrect Content-Type "%s"', $content_type)
-    if not exists $content_obj->{$content_type};
+  # TODO: support Content-Type: application/schema+json matching entry of application/json
 
-  if (exists $content_obj->{$content_type}{encoding}) {
-    my $state = { %$state, schema_path => jsonp($state->{schema_path}, 'content', $content_type) };
+  # we don't support */* here because what is the point? if you want to test string length,
+  # check the Content-Length header.
+  my $media_type = (first { $content_type eq fc } keys $content_obj->%*)
+    // (first { m{([^/]+)/\*$} && fc($content_type) =~ m{^\F$1/[^/]+$} } keys $content_obj->%*);
+  return E({ %$state, keyword => 'content' }, 'incorrect Content-Type "%s"', $content_type)
+    if not $media_type;
+
+  if (exists $content_obj->{$media_type}{encoding}) {
+    my $state = { %$state, schema_path => jsonp($state->{schema_path}, 'content', $media_type) };
     # "The key, being the property name, MUST exist in the schema as a property."
-    foreach my $property (sort keys $content_obj->{$content_type}{encoding}->%*) {
+    foreach my $property (sort keys $content_obj->{$media_type}{encoding}->%*) {
       ()= E({ $state, schema_path => jsonp($state->{schema_path}, 'schema', 'properties', $property) },
           'encoding property "%s" requires a matching property definition in the schema')
-        if not exists(($content_obj->{$content_type}{schema}{properties}//{})->{$property});
+        if not exists(($content_obj->{$media_type}{schema}{properties}//{})->{$property});
     }
 
     # "The encoding object SHALL only apply to requestBody objects when the media type is multipart or
@@ -352,26 +397,33 @@ sub _validate_body_content ($self, $state, $content_obj, $message) {
         \ Encode::decode($charset, $decoded_content_ref->$*, Encode::FB_CROAK | Encode::LEAVE_SRC);
     }
     catch ($e) {
-      return E({ %$state, keyword => 'content', _schema_path_suffix => $content_type },
+      return E({ %$state, keyword => 'content', _schema_path_suffix => $media_type },
         'could not decode content as %s', $charset);
     }
   }
 
-  my $media_type_decoder = $self->get_media_type($content_type);
-  abort({ %$state, keyword => 'content', _schema_path_suffix => $content_type },
+  my $schema = $content_obj->{$media_type}{schema};
+
+  # use the original Content-Type, NOT the possibly wildcard media type from the document
+  my $media_type_decoder = $self->get_media_type($content_type);  # case-insensitive, wildcard lookup
+  if (not $media_type_decoder) {
+    # don't fail if the schema would pass on any input
+    return if is_plain_hashref($schema) ? !keys %$schema : $schema;
+
+    abort({ %$state, keyword => 'content', _schema_path_suffix => $media_type },
       'EXCEPTION: unsupported Content-Type "%s": add support with $openapi->add_media_type(...)', $content_type)
-    if not $media_type_decoder;
+  }
 
   try {
     $decoded_content_ref = $media_type_decoder->($decoded_content_ref);
   }
   catch ($e) {
-    return E({ %$state, keyword => 'content', _schema_path_suffix => $content_type },
-      'could not decode content as %s', $content_type);
+    return E({ %$state, keyword => 'content', _schema_path_suffix => $media_type },
+      'could not decode content as %s', $media_type );
   }
 
-  $state = { %$state, schema_path => jsonp($state->{schema_path}, 'content', $content_type, 'schema') };
-  $self->_evaluate_subschema($decoded_content_ref->$*, $content_obj->{$content_type}{schema}, $state);
+  $state = { %$state, schema_path => jsonp($state->{schema_path}, 'content', $media_type, 'schema') };
+  $self->_evaluate_subschema($decoded_content_ref->$*, $schema, $state);
 }
 
 # wrap a result object around the errors
@@ -445,7 +497,7 @@ OpenAPI::Modern - Validate HTTP requests and responses against an OpenAPI docume
 
 =head1 VERSION
 
-version 0.008
+version 0.009
 
 =head1 SYNOPSIS
 
@@ -465,6 +517,7 @@ version 0.008
         schema:
           pattern: ^[a-z]+$
       post:
+        operationId: my_foo_request
         parameters:
         - name: My-Request-Header
           in: header
@@ -517,7 +570,7 @@ version 0.008
     [ 'My-Response-Header' => '123' ],
     '{"status": "ok"}',
   );
-  say $openapi->validate_response($response, '/foo/{foo_id}');
+  say $openapi->validate_response($response, { operation_id => 'my_foo_request' });
 
 prints:
 
@@ -586,9 +639,14 @@ The L<JSON::Schema::Modern> object to use for all URI resolution and JSON Schema
   $result = $openapi->validate_request(
     $request,
     {
-      path_captures => { arg1 => 1, arg2 => 2 },
       path_template => '/foo/{arg1}/bar/{arg2}',
+      path_captures => { arg1 => 1, arg2 => 2 },
     },
+    # OR:
+    # {
+    #   operation_id => 'my_operation_id',
+    #   path_captures => { arg1 => 1, arg2 => 2 },
+    # },
   );
 
 Validates an L<HTTP::Request> object against the corresponding OpenAPI v3.1 document, returning a
@@ -604,11 +662,17 @@ C<path_template>: a string representing the request URI, with placeholders in br
 
 =item *
 
+C<operation_id>: a string corresponding to the C<operationId> at a particular path-template and HTTP location under C</paths>
+
+=item *
+
 C<path_captures>: a hashref mapping placeholders in the path to their actual values in the request URI
 
 =back
 
 More options will be added later, providing more flexible matching of the document to the request.
+C<path_template> OR C<operation_id> is required.
+C<path_captures> is required.
 
 =head2 validate_response
 
@@ -630,6 +694,23 @@ C<path_template>: a string representing the request URI, with placeholders in br
 =back
 
 More options will be added later, providing more flexible matching of the document to the request.
+C<path_template> OR C<operation_id> is required.
+
+=head2 canonical_uri
+
+An accessor that delegates to L<JSON::Schema::Modern::Document/canonical_uri>.
+
+-head2 schema
+
+An accessor that delegates to L<JSON::Schema::Modern::Document/schema>.
+
+=head2 get_media_type
+
+An accessor that delegates to L<JSON::Schema::Modern/get_media_type>.
+
+=head2 add_media_type
+
+A setter that delegates to L<JSON::Schema::Modern/add_media_type>.
 
 =head1 ON THE USE OF JSON SCHEMAS
 
@@ -676,10 +757,6 @@ for query parameters, only C<style: form> is supported
 
 =item *
 
-for header parameters, only C<style: simple> is supported
-
-=item *
-
 cookie parameters are not checked at all yet
 
 =item *
